#include <tchar.h>
#include <windows.h>
#include <stdlib.h>
#include "../../litestep/lsapi/lsapi.h"
#include "jkey.h"

LPCTSTR szAppClass = _T("JKEY_CLASS"); // Class Name
LPCTSTR szAppName = _T(""); // Window Name

LPCTSTR buildName = _T("jkey.dll"); // Build (module) Name
LPCTSTR buildRev = _T("0.32"); // Build Version
LPCTSTR buildDate = _T("2000.10.19"); // Build Date
LPCTSTR buildAuthor = _T("jugg"); // Who last modified

LRESULT CALLBACK WndProc( HWND, _TINT, WPARAM, LPARAM );
void jKeyErrorBox( jKeyType*, LPCTSTR );
void freejKeys();
void loadjKeys();
void loadVKTable( LPTSTR );
BOOL stripWS ( LPTSTR );

HWND hMainWnd;
HWND hParentWnd;
HINSTANCE hInstance;

ATOM LWinKey;
ATOM RWinKey;

_TINT numvKeys = 0;
_TINT numjKeys = 0;

tVKTable *vkTable = NULL;
jKeyType *jkeys = NULL;
jkSettings jks;

int initModuleEx( HWND parent, HINSTANCE dllInst, const char* szPath )
{
	szPath=szPath;
	hInstance = dllInst;
	hParentWnd = parent;

	{	// Register the jKey window class
		WNDCLASS wc;
		memset(&wc, 0, sizeof(wc));
		wc.lpfnWndProc = (WNDPROC)WndProc;
		wc.hInstance = hInstance;
		wc.lpszClassName = szAppClass;
		wc.style = 0;
		if (!RegisterClass(&wc)){
			MessageBox(NULL,_T("Error registering jKey window class"),buildName,MB_OK|MB_ICONERROR|MB_DEFBUTTON1|MB_SYSTEMMODAL);
			return 1;
		}
	}

		// Create jKey window
	hMainWnd = CreateWindowEx(WS_EX_TOOLWINDOW,szAppClass,szAppName,WS_CHILD,0,0,0,0,hParentWnd,NULL,hInstance,NULL);
	if (hMainWnd == NULL){
		MessageBox(NULL,_T("Error creating jKey window!"),buildName,MB_OK|MB_ICONERROR|MB_DEFBUTTON1|MB_SYSTEMMODAL);
		UnregisterClass(szAppClass, dllInst);
		return 1;
	}

		// Get jKey settings
	jks.bNoWarn = GetRCBool(_T("jKeyNoWarn"),TRUE);
	jks.bUseHotkeyDef = GetRCBool(_T("jKeyUseHotkeyDef"),TRUE);
	jks.nLWinKeyTimeout = GetRCInt(_T("jKeyLWinKeyTimeout"), 750);
	jks.nRWinKeyTimeout = GetRCInt(_T("jKeyRWinKeyTimeout"), 750);

	{	// Load extended key items
		_TCHAR tmp[4096];

		*tmp = 0;
		GetRCLine(_T("jKeyVKTable"), tmp, 4096, _T(""));
		if (*tmp){
			jks.pVKTable = (LPTSTR)malloc((_tcslen(tmp)+1)*sizeof(_TCHAR));
			if (jks.pVKTable != NULL){
				_tcscpy(jks.pVKTable, tmp);
				loadVKTable(jks.pVKTable);
			}
		}

		*tmp = 0;
		GetRCLine(_T("jKeyLWinKey"), tmp, 4096, _T(""));
		if (*tmp){
			jks.pLWinKey = (LPTSTR)malloc((_tcslen(tmp)+1)*sizeof(_TCHAR));
			if (jks.pLWinKey != NULL){
				_tcscpy(jks.pLWinKey, tmp);
				if ((LWinKey=AddAtom(_T("LWIN_KEY"))) != 0){
					if (!RegisterHotKey(hMainWnd, LWinKey, MOD_WIN, VK_LWIN)){
						DeleteAtom(LWinKey);
						LWinKey = 0;
						if (!jks.bNoWarn)
							MessageBox(NULL,_T("Error registering Left WinKey!"),buildName,MB_OK|MB_ICONERROR|MB_DEFBUTTON1|MB_SYSTEMMODAL);
					}
				}
				else if (!jks.bNoWarn)
					MessageBox(NULL,_T("Error registering Left WinKey!"),buildName,MB_OK|MB_ICONERROR|MB_DEFBUTTON1|MB_SYSTEMMODAL);
			}
		}else
			jks.pLWinKey = NULL;

		*tmp = 0;
		GetRCLine(_T("jKeyRWinKey"), tmp, 4096, _T(""));
		if (*tmp){
			jks.pRWinKey = (LPTSTR)malloc((_tcslen(tmp)+1)*sizeof(_TCHAR));
			if (jks.pRWinKey != NULL){
				_tcscpy(jks.pRWinKey, tmp);
				if ((RWinKey=AddAtom(_T("RWIN_KEY"))) != 0){
					if (!RegisterHotKey(hMainWnd, RWinKey, MOD_WIN, VK_RWIN)){
						DeleteAtom(RWinKey);
						RWinKey = 0;
						if (!jks.bNoWarn)
							MessageBox(NULL,_T("Error registering Right WinKey!"),buildName,MB_OK|MB_ICONERROR|MB_DEFBUTTON1|MB_SYSTEMMODAL);
					}
				}
				else if (!jks.bNoWarn)
					MessageBox(NULL,_T("Error registering Right WinKey!"),buildName,MB_OK|MB_ICONERROR|MB_DEFBUTTON1|MB_SYSTEMMODAL);
			}
		}else
			jks.pRWinKey = NULL;
	}

		// Get jKey definitions
	loadjKeys();

	{	//register message for version info
		UINT Msgs[2];
		Msgs[0] = LM_GETREVID;
		Msgs[1] = 0;
		SendMessage(hParentWnd, LM_REGISTERMESSAGE, (WPARAM)hMainWnd, (LPARAM)Msgs);
	}
	return 0;
}

void quitModule( HINSTANCE dllInst )
{
	{	//unregister message for version info
		UINT Msgs[2];
		Msgs[0] = LM_GETREVID;
		Msgs[1] = 0;
		SendMessage(hParentWnd, LM_UNREGISTERMESSAGE, (WPARAM)hMainWnd, (LPARAM)Msgs);
	}

	if (LWinKey)
		UnregisterHotKey(hMainWnd, LWinKey);
	if (jks.pLWinKey != NULL)
		free(jks.pLWinKey);

	if (RWinKey)
		UnregisterHotKey(hMainWnd, RWinKey);
	if (jks.pRWinKey != NULL)
		free(jks.pRWinKey);

	freejKeys();

	if (jks.pVKTable != NULL){
		_TINT i;
		for (i=0;i<numvKeys;i++){
			free(vkTable[i].key);
			vkTable[i].key = NULL;
		}
		if (vkTable != NULL){
			free(vkTable);
			vkTable = NULL;
		}
	}

	if (hMainWnd != NULL)
		DestroyWindow(hMainWnd);

	UnregisterClass(szAppClass, dllInst);
	return;
}

LRESULT CALLBACK WndProc( HWND hwnd, _TINT message, WPARAM wParam, LPARAM lParam )
{
	switch (message){
		case LM_GETREVID:
			if (wParam == 0)
				_stprintf((LPTSTR)lParam, _T("%s: %s"), buildName, buildRev);
			else if (wParam == 1)
				_stprintf((LPTSTR)lParam, _T("%s: %s - %s - %s"), buildName, buildRev, buildDate, buildAuthor);
			else
				_tcscpy((LPTSTR)lParam, _T(""));
			return _tcslen((LPTSTR)lParam);

		case WM_CLOSE:
			return 0;

		case WM_SYSCOMMAND:
			switch (wParam){
				case SC_CLOSE:
					PostMessage(hParentWnd, LM_RECYCLE, 3, 0);
					return 0;
				default:
					break;
			}
			break;

		case WM_DESTROY:
			hMainWnd = NULL;
			return 0;

		case WM_HOTKEY:
			if ((_TINT)wParam < numjKeys){
				KillTimer(hMainWnd, 1);
				KillTimer(hMainWnd, 2);
				LSExecute(hMainWnd, jkeys[(_TINT)wParam].cmd, 0);
			}
			else if (LWinKey && (ATOM)wParam == LWinKey){
				KillTimer(hMainWnd, 1);
				SetTimer(hMainWnd, 1, jks.nLWinKeyTimeout, NULL);
			}
			else if (RWinKey && (ATOM)wParam == RWinKey){
				KillTimer(hMainWnd, 2);
				SetTimer(hMainWnd, 2, jks.nRWinKeyTimeout, NULL);
			}
			break;

		case WM_TIMER:
			if (wParam == 1){
				KillTimer(hMainWnd, 1);
				LSExecute(hMainWnd, jks.pLWinKey, 0);
			}
			else if (wParam == 2){
				KillTimer(hMainWnd, 2);
				LSExecute(hMainWnd, jks.pRWinKey, 0);
			}
			return 0;

		default:
			break;
	}
	return DefWindowProc(hwnd,message,wParam,lParam);
}

void jKeyErrorBox( jKeyType* jKey, LPCTSTR tmpkey )
{
	_TCHAR tmp0[128];
	_TCHAR tmp1[128];
	_TCHAR tmp2[4096];
	_TINT i;
	_TINT len;

	*tmp0 = *tmp1 = *tmp2 = 0;
	i = len = 0;

	if ((jKey->modkey & MOD_ALT) == MOD_ALT){
		_tcscpy(tmp0, _T("ALT"));
		len = 3;
	}
	if ((jKey->modkey & MOD_CONTROL) == MOD_CONTROL){
		if (len > 0){
			_tcscat(tmp0, _T("+"));
			len += 1;
		}
		_tcscat(tmp0, _T("CTRL"));
		len += 4;
	}
	if ((jKey->modkey & MOD_SHIFT) == MOD_SHIFT){
		if (len > 0){
			_tcscat(tmp0, _T("+"));
			len += 1;
		}
		_tcscat(tmp0, _T("SHIFT"));
		len += 5;
	}
	if ((jKey->modkey & MOD_WIN) == MOD_WIN){
		if (len > 0){
			_tcscat(tmp0, _T("+"));
			len += 1;
		}
		_tcscat(tmp0, _T("WIN"));
		len += 3;
	}
	tmp0[len] = 0;

	if (!_istalnum(jKey->ckey)){
		if (tmpkey != NULL)
			_tcscpy(tmp1, tmpkey);
		else{
			for (i=0; i<numvKeys; i++){
				if (vkTable[i].vkey == jKey->ckey){
					_tcscpy(tmp1, vkTable[i].key);
					i = numvKeys;
				}
			}
		}
	}else{
		tmp1[0] = jKey->ckey;
		tmp1[1] = 0;
	}

	_stprintf(tmp2, _T("Error registering %s %s %s %s"), (jks.bUseHotkeyDef ? _T("*Hotkey"):_T("*jKey")), tmp0, tmp1, jKey->cmd);

	MessageBox(NULL,tmp2,buildName,MB_OK|MB_ICONERROR|MB_DEFBUTTON1|MB_SYSTEMMODAL);
	return;
}

void freejKeys()
{
	_TINT i;
	for (i=0;i<numjKeys;i++){
		UnregisterHotKey(hMainWnd,i);
		if (jkeys[i].cmd != NULL){
			free(jkeys[i].cmd);
			jkeys[i].cmd = NULL;
		}
	}
	if (jkeys != NULL){
		free(jkeys);
		jkeys = NULL;
	}
	return;
}

void loadjKeys()
{
	FILE *f = LCOpen(NULL);
	if (f){
		_TCHAR buffer[4096];
		_TCHAR token0[128], token1[128], token2[128], cmdstr[4096];
		LPTSTR tokens[3];
		_TINT count;
		jKeyType tmpjKey;

		tokens[0] = token0;
		tokens[1] = token1;
		tokens[2] = token2;

		while (LCReadNextConfig(f, jks.bUseHotkeyDef ? _T("*HotKey"):_T("*jKey"), buffer, sizeof(buffer))){
			if ((count = LCTokenize(buffer, tokens, 3, cmdstr)) == 3){
				if (stripWS(token1) && stripWS(token2) && stripWS(cmdstr)){
					memset(&tmpjKey, 0, sizeof(jKeyType));

					tokens[1] = _tcstok(token1, _T("+"));
					while (tokens[1] != NULL){
						if (!_tcsicmp(tokens[1], _T("Win")))
							tmpjKey.modkey |= MOD_WIN;
						else if (!_tcsicmp(tokens[1], _T("Ctrl")))
							tmpjKey.modkey |= MOD_CONTROL;
						else if (!_tcsicmp(tokens[1], _T("Alt")))
							tmpjKey.modkey |= MOD_ALT;
						else if (!_tcsicmp(tokens[1], _T("Shift")))
							tmpjKey.modkey |= MOD_SHIFT;
						tokens[1] = _tcstok(NULL, _T("+"));
					}
					tokens[1] = token1; // move pointer back to beginning of token1

					if (_tcslen(token2) == 1)
						tmpjKey.ckey = *token2;
					else{
						for (count=0; count<numvKeys && !tmpjKey.ckey; count++){
							if (!_tcsicmp(vkTable[count].key, token2))
								tmpjKey.ckey = vkTable[count].vkey;
						}
						// even if no 'vkey' was found, letting it fall through here,
						// as it will fail on RegisterHotKey. Not the most efficient
						// thing, but easiest for now, plus it shouldn't happen often
						// once the user gets their hotkeys setup correctly.
					}

					tmpjKey.cmd = (LPTSTR)malloc((_tcslen(cmdstr)+1)*sizeof(_TCHAR));
					if (tmpjKey.cmd != NULL){
						_tcscpy(tmpjKey.cmd, cmdstr);

						if (!RegisterHotKey(hMainWnd,numjKeys,tmpjKey.modkey,tmpjKey.ckey)){
							if (!jks.bNoWarn)
								jKeyErrorBox(&tmpjKey, (count == numvKeys ? token2:NULL));
							free(tmpjKey.cmd);
							tmpjKey.cmd = NULL;
						}else{
							jKeyType * tempKeys = jkeys;
							if (!jkeys)
								jkeys = (jKeyType *)malloc(sizeof(jKeyType));
							else
								jkeys = (jKeyType *)realloc(jkeys, (numjKeys+1)*sizeof(jKeyType));
							if (jkeys == NULL){
								jkeys = tempKeys;
								free(tmpjKey.cmd);
								tmpjKey.cmd = NULL;
							}else{
								memmove(&jkeys[numjKeys], &tmpjKey, sizeof(jKeyType));
								++numjKeys;
							}
							tempKeys = NULL;
						}
					}
				}
			}
		}
		LCClose(f);
	}
	return;
}

void loadVKTable( LPTSTR table )
{
	FILE *f;
	tVKTable tmpTable;
	_TCHAR tmpBuf[128];
	_TCHAR tBuf[128];
	_TINT nKey;

	f = _tfopen(table, _T("r"));
	if (f != NULL)
		fseek(f, 0, SEEK_SET);
	else
		return;

	numvKeys = 0;
	while (f && !ferror(f) && !feof(f)){
		if (!_fgetts(tmpBuf, 256, f))
			continue;
		if (stripWS(_tcsupr(tmpBuf)) && *tmpBuf != _T(';')){
			if (_stscanf(tmpBuf, _T("%s , %X"), tBuf, &nKey)){
				memset(&tmpTable, 0, sizeof(tVKTable));
				tmpTable.key = (LPTSTR)malloc((_tcslen(tBuf)+1)*sizeof(_TCHAR));
				if (tmpTable.key != NULL){
					tVKTable * tempVKT = vkTable;

					_tcscpy(tmpTable.key, tBuf);
					tmpTable.vkey = (_TCHAR)nKey;
 
					if (!vkTable)
						vkTable = (tVKTable *)malloc(sizeof(tVKTable));
					else
						vkTable = (tVKTable *)realloc(vkTable, (numvKeys+1)*sizeof(tVKTable));
					if (vkTable == NULL){
						vkTable = tempVKT;
						free (tmpTable.key);
						tmpTable.key = NULL;
					}else{
						memmove(&vkTable[numvKeys], &tmpTable, sizeof(tVKTable));
						++numvKeys;
					}
				}
			}
		}
	}
	fclose(f);
	return;
}

BOOL stripWS ( LPTSTR buffer )
{
	_TINT length = (*buffer && buffer != NULL) ? _tcslen(buffer):0;
	if (length){
		while (length && _istspace(*buffer)){
			_tcscpy(buffer, buffer + 1);
			--length;
		}
		while (length && _istspace(buffer[length-1]))
			buffer[--length] = 0;
		if (length) return TRUE;
		if (buffer != NULL)
			*buffer = 0;
	}
	return FALSE;
}
